#include <stdio.h>
#include <dos.h>

/*declarations imported from header*/
#include "comms.h"

/*probably would be better to retrieve LPT port addy...*/
unsigned int LPT1 = 0x378;

/*****************************************
Generic comms functions courtesy of hcs
http://sourceforge.net/projects/N64dev
nifty, eh?
*****************************************/

/*call to initialize comms
This locks the game and allows PC access
You must re-initialize between read/write to reset pipe*/
int InitGSComms() {
	int timeout=0x60;/*change this to fiddle with general timeout value*/
	while (GSFcn1(3) && timeout) {
		timeout--;
		/*15 second warning...*/
		if(timeout==0x40) printf("\ntimeout in:   %2d",timeout);
		if(timeout<=0x40) printf("\b\b%2d",timeout);
	}
	if (!timeout) erratta(1);
	if (!CheckGSPresence()) erratta(2);
    return 0;
}

/*closing comms cuts the stream.
You only really need to do it when all access is complete
Therefore, read&write all day long, then kill stream at the very end*/
int CloseGSComms(void) {
	if (GSFcn1(3)|(!CheckGSPresence())) erratta(5);
	ReadWriteByte(0x64);
    outport(LPT1,0);      /*blank out the LPT*/
    return 0;
}

/*checks for sync*/
int GSFcn1(int x) {
	int timeout=0x28;
	unsigned char result;

	while (timeout) {
		result<<=4;
		result|=SendNibble(x);
		/* when we recieve a 6, then a 7, we're sync'd*/
		if (result==0x67) break;

		SendNibble(x);
		SendNibble(x>>4);

		timeout--;
	}

	if (!timeout) return 1;
	return 0;
}

int SendNibble(int x) {
	int timeout=0x3e8,retval;

	outport(LPT1,(x & 0xf)|0x10);

	while (timeout && (inport(LPT1+1) & 8)) timeout--;

	retval=((inport(LPT1+1)& 0xf0)^0x80)>>4;

	outport(LPT1,0);

	timeout=0x64;
	while (timeout && (inport(LPT1+1)& 8)) timeout--;

	if (!timeout) return 0;
	else return retval;
}

int ReadWriteNibble(int x) {
	int retval;
	outport(LPT1,(x & 0xf)|0x10);

	while ((~inport(LPT1+1))&8);
	while ((~inport(LPT1+1))&8);

	retval=((inport(LPT1+1)& 0xf0)^0x80)>>4;

	outport(LPT1,0);

	while (inport(LPT1+1)& 8);
	while (inport(LPT1+1)& 8);

	return retval;
}

int ReadWriteByte(int x) {
	return (ReadWriteNibble(x>>4)<<4)|ReadWriteNibble(x);
}

unsigned int ReadWrite16(unsigned int x) {
	return ((ReadWriteNibble(x>>12)<<12)|(ReadWriteNibble(x>>8)<<8)|(ReadWriteNibble(x>>4)<<4)|ReadWriteNibble(x));
}

unsigned long ReadWrite32(unsigned long x) {
	return ((unsigned long)ReadWriteNibble(x>>28)<<28)|\
	((unsigned long)ReadWriteNibble(x>>24)<<24)|\
	((unsigned long)ReadWriteNibble(x>>20)<<20)|\
	((unsigned long)ReadWriteNibble(x>>16)<<16)|\
	((unsigned long)ReadWriteNibble(x>>12)<<12)|\
	((unsigned long)ReadWriteNibble(x>>8)<<8)|\
	((unsigned long)ReadWriteNibble(x>>4)<<4)|\
	(unsigned long)ReadWriteNibble(x);
}


/* this breaks when the GS can allow ram access*/
int CheckGSPresence(void) {
	int timeout=0x3e8,result;

	while (timeout) {
		result=(ReadWriteNibble(0x47>>4)<<4)|ReadWriteNibble(0x47);
		timeout--;
		if (result != 0x67) continue;
		result=(ReadWriteNibble(0x54>>4)<<4)|ReadWriteNibble(0x54);
		if (result == 0x74) break;
	}

/*	printf("\nGS utilities can't be used while in the GS menu.\n");*/
	if (timeout) return 1;
	return 0;
}

/*these are simple read/write types as above, but don't initiate or close the comms
  These are slightly faster, but the others are more robust*/
unsigned long Read32(unsigned long address) {
	ReadWrite32(address);
	ReadWrite32(4);
	return ReadWrite32(0);
}


unsigned short Read16(unsigned long address) {
	ReadWrite32(address);
	ReadWrite32(2);
	return ReadWrite16(0);
}

unsigned char ReadByte(unsigned long address)
{ReadWrite32(address);
ReadWrite32(1);
return ReadWriteByte(0);
}


int Write32(unsigned long address,unsigned long value) {
	ReadWrite32(address);
	ReadWrite32(4);
	ReadWrite32(value);
	return 0;
}

int Write16(unsigned long address,unsigned short value) {
	ReadWrite32(address);
	ReadWrite32(2);
	ReadWrite16(value);
	return 0;
}

int WriteByte(unsigned long address, unsigned char value) {
	ReadWrite32(address);
	ReadWrite32(1);
	ReadWriteByte(value);
	return 0;
}

/*File Upload*/
int UploadFile(FILE * infile, unsigned long address) {
	unsigned long c=0;
	unsigned long size;
	char buf;

	if (fseek(infile,0,SEEK_END)) return 1;
	size=ftell(infile);
	rewind(infile);

	if (InitGSComms()) return 1;

	printf("Uploading...00%%");

	ReadWriteByte(2);
	ReadWrite32(address);
	ReadWrite32(size);

	for (c=0; c < size; c++) {
		fread(&buf,1,1,infile);
		if (c & 0x400) printf("\b\b\b%2d%%",c*100/size);
		ReadWriteByte(buf);
	}

	ReadWrite32(0);
	ReadWrite32(0);
	/*if (CloseGSComms()) return 1;*/

	printf("\b\b\bDone.\n");
	return 0;
}

/*Gameshark Code Generator List functions*/

/*grab #codes currently in list
returns #codes or -1 if error
initializes but does not close port*/

long NumActiveCode(void)
{long num;
if (InitGSComms()) return -1;
	ReadWriteByte(1);
	num=Read32(0xA07E9D60);
	ReadWrite32(0);
	ReadWrite32(0);
return num;}

/*add code to "active code" list
retrieve number of codes listed.
jump to point. (#codes * 6)
Send code, then update #

returns either number of codes in list or -1 if an error*/
int AddActiveCode(unsigned long addy, unsigned short val)
{unsigned long num;        /*number of codes in list*/

/*grab #codes currently in list*/
num=NumActiveCode();
if(num<0) return num;

/*write code*/
if (InitGSComms()) return -1;
   	ReadWriteByte(2);
    Write32(0xA07E9D64 + (num * 6),addy);
    Write16(0xA07E9D68 + (num * 6),val);
    Write32(0xA07E9D60,num + 1);
	ReadWrite32(0);
	ReadWrite32(0);

return num+1;}

int erratta(int code)
{if(code<1) return 0;
else if(code==1) fprintf(stderr,"\nTimeout!  Can not interface with I/O port.\n");
else if(code==2) fprintf(stderr,"\nGS menu detected.  Start normal play or exit the GS menus, then try again\n");
else if(code==5) fprintf(stderr,"\nUnable to close comms.\n");
else fprintf(stderr,"\nAborting...\n");
system("PAUSE");
system("CLS");
exit(code);
}